@@ -41,12 +41,11 @@
 zmq::curve_server_t::curve_server_t (session_base_t *session_,
                                      const std::string &peer_address_,
                                      const options_t &options_) :
-    mechanism_t (options_),
-    session (session_),
-    peer_address (peer_address_),
-    state (expect_hello),
-    cn_nonce (1),
-    cn_peer_nonce(1)
+    mechanism_base_t (session_, options_),
+    zap_client_common_handshake_t (
+      session_, peer_address_, options_, sending_ready),
+    curve_mechanism_base_t (
+      session_, options_, "CurveZMQMESSAGES", "CurveZMQMESSAGEC")
 {
     int rc;
     //  Fetch our secret key from socket options
@@ -66,17 +65,17 @@ int zmq::curve_server_t::next_handshake_command (msg_t *msg_)
     int rc = 0;
 
     switch (state) {
-        case send_welcome:
+        case sending_welcome:
             rc = produce_welcome (msg_);
             if (rc == 0)
-                state = expect_initiate;
+                state = waiting_for_initiate;
             break;
-        case send_ready:
+        case sending_ready:
             rc = produce_ready (msg_);
             if (rc == 0)
-                state = connected;
+                state = ready;
             break;
-        case send_error:
+        case sending_error:
             rc = produce_error (msg_);
             if (rc == 0)
                 state = error_sent;
@@ -94,15 +93,22 @@ int zmq::curve_server_t::process_handshake_command (msg_t *msg_)
     int rc = 0;
 
     switch (state) {
-        case expect_hello:
+        case waiting_for_hello:
             rc = process_hello (msg_);
             break;
-        case expect_initiate:
+        case waiting_for_initiate:
             rc = process_initiate (msg_);
             break;
         default:
-            //  Temporary support for security debugging
-            puts ("CURVE I: invalid handshake command");
+            // TODO I think this is not a case reachable with a misbehaving 
+            // client. It is not an "invalid handshake command", but would be 
+            // trying to process a handshake command in an invalid state, 
+            // which is purely under control of this peer. 
+            // Therefore, it should be changed to zmq_assert (false);
+
+            // CURVE I: invalid handshake command
+            session->get_socket ()->event_handshake_failed_protocol (
+              session->get_endpoint (), ZMQ_PROTOCOL_ERROR_ZMTP_UNSPECIFIED);
             errno = EPROTO;
             rc = -1;
             break;
@@ -118,165 +124,35 @@ int zmq::curve_server_t::process_handshake_command (msg_t *msg_)
 
 int zmq::curve_server_t::encode (msg_t *msg_)
 {
-    zmq_assert (state == connected);
-
-    const size_t mlen = crypto_box_ZEROBYTES + 1 + msg_->size ();
-
-    uint8_t message_nonce [crypto_box_NONCEBYTES];
-    memcpy (message_nonce, "CurveZMQMESSAGES", 16);
-    put_uint64 (message_nonce + 16, cn_nonce);
-
-    uint8_t flags = 0;
-    if (msg_->flags () & msg_t::more)
-        flags |= 0x01;
-    if (msg_->flags () & msg_t::command)
-        flags |= 0x02;
-
-    uint8_t *message_plaintext = static_cast <uint8_t *> (malloc (mlen));
-    alloc_assert (message_plaintext);
-
-    memset (message_plaintext, 0, crypto_box_ZEROBYTES);
-    message_plaintext [crypto_box_ZEROBYTES] = flags;
-    memcpy (message_plaintext + crypto_box_ZEROBYTES + 1,
-            msg_->data (), msg_->size ());
-
-    uint8_t *message_box = static_cast <uint8_t *> (malloc (mlen));
-    alloc_assert (message_box);
-
-    int rc = crypto_box_afternm (message_box, message_plaintext,
-                                 mlen, message_nonce, cn_precom);
-    zmq_assert (rc == 0);
-
-    rc = msg_->close ();
-    zmq_assert (rc == 0);
-
-    rc = msg_->init_size (16 + mlen - crypto_box_BOXZEROBYTES);
-    zmq_assert (rc == 0);
-
-    uint8_t *message = static_cast <uint8_t *> (msg_->data ());
-
-    memcpy (message, "\x07MESSAGE", 8);
-    memcpy (message + 8, message_nonce + 16, 8);
-    memcpy (message + 16, message_box + crypto_box_BOXZEROBYTES,
-            mlen - crypto_box_BOXZEROBYTES);
-
-    free (message_plaintext);
-    free (message_box);
-
-    cn_nonce++;
-
-    return 0;
+    zmq_assert (state == ready);
+    return curve_mechanism_base_t::encode (msg_);
 }
 
 int zmq::curve_server_t::decode (msg_t *msg_)
 {
-    zmq_assert (state == connected);
-
-    if (msg_->size () < 33) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: invalid CURVE client, sent malformed command");
-        errno = EPROTO;
-        return -1;
-    }
-
-    const uint8_t *message = static_cast <uint8_t *> (msg_->data ());
-    if (memcmp (message, "\x07MESSAGE", 8)) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: invalid CURVE client, did not send MESSAGE");
-        errno = EPROTO;
-        return -1;
-    }
-
-    uint8_t message_nonce [crypto_box_NONCEBYTES];
-    memcpy (message_nonce, "CurveZMQMESSAGEC", 16);
-    memcpy (message_nonce + 16, message + 8, 8);
-    uint64_t nonce = get_uint64(message + 8);
-    if (nonce <= cn_peer_nonce) {
-        errno = EPROTO;
-        return -1;
-    }
-    cn_peer_nonce = nonce;
-
-    const size_t clen = crypto_box_BOXZEROBYTES + msg_->size () - 16;
-
-    uint8_t *message_plaintext = static_cast <uint8_t *> (malloc (clen));
-    alloc_assert (message_plaintext);
-
-    uint8_t *message_box = static_cast <uint8_t *> (malloc (clen));
-    alloc_assert (message_box);
-
-    memset (message_box, 0, crypto_box_BOXZEROBYTES);
-    memcpy (message_box + crypto_box_BOXZEROBYTES,
-            message + 16, msg_->size () - 16);
-
-    int rc = crypto_box_open_afternm (message_plaintext, message_box,
-                                      clen, message_nonce, cn_precom);
-    if (rc == 0) {
-        rc = msg_->close ();
-        zmq_assert (rc == 0);
-
-        rc = msg_->init_size (clen - 1 - crypto_box_ZEROBYTES);
-        zmq_assert (rc == 0);
-
-        const uint8_t flags = message_plaintext [crypto_box_ZEROBYTES];
-        if (flags & 0x01)
-            msg_->set_flags (msg_t::more);
-        if (flags & 0x02)
-            msg_->set_flags (msg_t::command);
-
-        memcpy (msg_->data (),
-                message_plaintext + crypto_box_ZEROBYTES + 1,
-                msg_->size ());
-    }
-    else {
-        //  Temporary support for security debugging
-        puts ("CURVE I: connection key used for MESSAGE is wrong");
-        errno = EPROTO;
-    }
-    free (message_plaintext);
-    free (message_box);
-
-    return rc;
-}
-
-int zmq::curve_server_t::zap_msg_available ()
-{
-    if (state != expect_zap_reply) {
-        errno = EFSM;
-        return -1;
-    }
-    const int rc = receive_and_process_zap_reply ();
-    if (rc == 0)
-        state = status_code == "200"
-            ? send_ready
-            : send_error;
-    return rc;
-}
-
-zmq::mechanism_t::status_t zmq::curve_server_t::status () const
-{
-    if (state == connected)
-        return mechanism_t::ready;
-    else
-    if (state == error_sent)
-        return mechanism_t::error;
-    else
-        return mechanism_t::handshaking;
+    zmq_assert (state == ready);
+    return curve_mechanism_base_t::decode (msg_);
 }
 
 int zmq::curve_server_t::process_hello (msg_t *msg_)
 {
-    if (msg_->size () != 200) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: client HELLO is not correct size");
+    int rc = check_basic_command_structure (msg_);
+    if (rc == -1)
+      return -1;
+
+    const size_t size = msg_->size ();
+    const uint8_t * const hello = static_cast <uint8_t *> (msg_->data ());
+
+    if (size < 6 || memcmp (hello, "\x05HELLO", 6)) {
+        session->get_socket ()->event_handshake_failed_protocol (
+          session->get_endpoint (), ZMQ_PROTOCOL_ERROR_ZMTP_UNEXPECTED_COMMAND);
         errno = EPROTO;
         return -1;
     }
 
-    const uint8_t * const hello = static_cast <uint8_t *> (msg_->data ());
-    if (memcmp (hello, "\x05HELLO", 6)) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: client HELLO has invalid command name");
+    if (size != 200) {
+        session->get_socket ()->event_handshake_failed_protocol (
+          session->get_endpoint (), ZMQ_PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_HELLO);
         errno = EPROTO;
         return -1;
     }
@@ -285,8 +161,9 @@ int zmq::curve_server_t::process_hello (msg_t *msg_)
     const uint8_t minor = hello [7];
 
     if (major != 1 || minor != 0) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: client HELLO has unknown version number");
+        // CURVE I: client HELLO has unknown version number
+        session->get_socket ()->event_handshake_failed_protocol (
+          session->get_endpoint (), ZMQ_PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_HELLO);
         errno = EPROTO;
         return -1;
     }
@@ -306,17 +183,17 @@ int zmq::curve_server_t::process_hello (msg_t *msg_)
     memcpy (hello_box + crypto_box_BOXZEROBYTES, hello + 120, 80);
 
     //  Open Box [64 * %x0](C'->S)
-    int rc = crypto_box_open (hello_plaintext, hello_box,
-                              sizeof hello_box,
-                              hello_nonce, cn_client, secret_key);
+    rc = crypto_box_open (hello_plaintext, hello_box, sizeof hello_box,
+                          hello_nonce, cn_client, secret_key);
     if (rc != 0) {
-        //  Temporary support for security debugging
-        puts("CURVE I: cannot open client HELLO -- wrong server key?");
+        // CURVE I: cannot open client HELLO -- wrong server key?
+        session->get_socket ()->event_handshake_failed_protocol (
+          session->get_endpoint (), ZMQ_PROTOCOL_ERROR_ZMTP_CRYPTOGRAPHIC);
         errno = EPROTO;
         return -1;
     }
-    
-    state = send_welcome;
+
+    state = sending_welcome;
     return rc;
 }
 
@@ -365,8 +242,15 @@ int zmq::curve_server_t::produce_welcome (msg_t *msg_)
             cookie_ciphertext + crypto_secretbox_BOXZEROBYTES, 80);
 
     rc = crypto_box (welcome_ciphertext, welcome_plaintext,
-                     sizeof welcome_plaintext,
-                     welcome_nonce, cn_client, secret_key);
+                     sizeof welcome_plaintext, welcome_nonce, cn_client,
+                     secret_key);
+
+    //  TODO I think we should change this back to zmq_assert (rc == 0);
+    //  as it was before https://github.com/zeromq/libzmq/pull/1832
+    //  The reason given there was that secret_key might be 0ed.
+    //  But if it were, we would never get this far, since we could
+    //  not have opened the client's hello box with a 0ed key.
+
     if (rc == -1)
         return -1;
 
@@ -383,17 +267,24 @@ int zmq::curve_server_t::produce_welcome (msg_t *msg_)
 
 int zmq::curve_server_t::process_initiate (msg_t *msg_)
 {
-    if (msg_->size () < 257) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: client INITIATE is not correct size");
+    int rc = check_basic_command_structure (msg_);
+    if (rc == -1)
+      return -1;
+
+    const size_t size = msg_->size ();
+    const uint8_t *initiate = static_cast <uint8_t *> (msg_->data ());
+
+    if (size < 9 || memcmp (initiate, "\x08INITIATE", 9)) {
+        session->get_socket ()->event_handshake_failed_protocol (
+          session->get_endpoint (), ZMQ_PROTOCOL_ERROR_ZMTP_UNEXPECTED_COMMAND);
         errno = EPROTO;
         return -1;
     }
 
-    const uint8_t *initiate = static_cast <uint8_t *> (msg_->data ());
-    if (memcmp (initiate, "\x08INITIATE", 9)) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: client INITIATE has invalid command name");
+    if (size < 257) {
+        session->get_socket ()->event_handshake_failed_protocol (
+          session->get_endpoint (),
+          ZMQ_PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_INITIATE);
         errno = EPROTO;
         return -1;
     }
@@ -409,12 +300,12 @@ int zmq::curve_server_t::process_initiate (msg_t *msg_)
     memcpy (cookie_nonce, "COOKIE--", 8);
     memcpy (cookie_nonce + 8, initiate + 9, 16);
 
-    int rc = crypto_secretbox_open (cookie_plaintext, cookie_box,
-                                    sizeof cookie_box,
-                                    cookie_nonce, cookie_key);
+    rc = crypto_secretbox_open (cookie_plaintext, cookie_box, sizeof cookie_box,
+                                cookie_nonce, cookie_key);
     if (rc != 0) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: cannot open client INITIATE cookie");
+        // CURVE I: cannot open client INITIATE cookie
+        session->get_socket ()->event_handshake_failed_protocol (
+          session->get_endpoint (), ZMQ_PROTOCOL_ERROR_ZMTP_CRYPTOGRAPHIC);
         errno = EPROTO;
         return -1;
     }
@@ -422,13 +313,17 @@ int zmq::curve_server_t::process_initiate (msg_t *msg_)
     //  Check cookie plain text is as expected [C' + s']
     if (memcmp (cookie_plaintext + crypto_secretbox_ZEROBYTES, cn_client, 32)
     ||  memcmp (cookie_plaintext + crypto_secretbox_ZEROBYTES + 32, cn_secret, 32)) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: client INITIATE cookie is not valid");
+        // TODO this case is very hard to test, as it would require a modified
+        //  client that knows the server's secret temporary cookie key
+
+        // CURVE I: client INITIATE cookie is not valid
+        session->get_socket ()->event_handshake_failed_protocol (
+          session->get_endpoint (), ZMQ_PROTOCOL_ERROR_ZMTP_CRYPTOGRAPHIC);
         errno = EPROTO;
         return -1;
     }
 
-    const size_t clen = (msg_->size () - 113) + crypto_box_BOXZEROBYTES;
+    const size_t clen = (size - 113) + crypto_box_BOXZEROBYTES;
 
     uint8_t initiate_nonce [crypto_box_NONCEBYTES];
     uint8_t initiate_plaintext [crypto_box_ZEROBYTES + 128 + 256];
@@ -446,8 +341,9 @@ int zmq::curve_server_t::process_initiate (msg_t *msg_)
     rc = crypto_box_open (initiate_plaintext, initiate_box,
                           clen, initiate_nonce, cn_client, cn_secret);
     if (rc != 0) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: cannot open client INITIATE");
+        // CURVE I: cannot open client INITIATE
+        session->get_socket ()->event_handshake_failed_protocol (
+          session->get_endpoint (), ZMQ_PROTOCOL_ERROR_ZMTP_CRYPTOGRAPHIC);
         errno = EPROTO;
         return -1;
     }
@@ -471,16 +367,21 @@ int zmq::curve_server_t::process_initiate (msg_t *msg_)
                           sizeof vouch_box,
                           vouch_nonce, client_key, cn_secret);
     if (rc != 0) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: cannot open client INITIATE vouch");
+        // CURVE I: cannot open client INITIATE vouch
+        session->get_socket ()->event_handshake_failed_protocol (
+          session->get_endpoint (), ZMQ_PROTOCOL_ERROR_ZMTP_CRYPTOGRAPHIC);
         errno = EPROTO;
         return -1;
     }
 
     //  What we decrypted must be the client's short-term public key
     if (memcmp (vouch_plaintext + crypto_box_ZEROBYTES, cn_client, 32)) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: invalid handshake from client (public key)");
+        // TODO this case is very hard to test, as it would require a modified
+        //  client that knows the server's secret short-term key
+
+        // CURVE I: invalid handshake from client (public key)
+        session->get_socket ()->event_handshake_failed_protocol (
+          session->get_endpoint (), ZMQ_PROTOCOL_ERROR_ZMTP_KEY_EXCHANGE);
         errno = EPROTO;
         return -1;
     }
@@ -489,23 +390,35 @@ int zmq::curve_server_t::process_initiate (msg_t *msg_)
     rc = crypto_box_beforenm (cn_precom, cn_client, cn_secret);
     zmq_assert (rc == 0);
 
-    //  Use ZAP protocol (RFC 27) to authenticate the user.
-    rc = session->zap_connect ();
-    if (rc == 0) {
-        send_zap_request (client_key);
-        rc = receive_and_process_zap_reply ();
-        if (rc == 0)
-            state = status_code == "200"
-                ? send_ready
-                : send_error;
-        else
-        if (errno == EAGAIN)
-            state = expect_zap_reply;
-        else
+    //  Given this is a backward-incompatible change, it's behind a socket
+    //  option disabled by default.
+    if (zap_required () || !options.zap_enforce_domain) {
+        //  Use ZAP protocol (RFC 27) to authenticate the user.
+        rc = session->zap_connect ();
+        if (rc == 0) {
+            send_zap_request (client_key);
+            state = waiting_for_zap_reply;
+
+            //  TODO actually, it is quite unlikely that we can read the ZAP 
+            //  reply already, but removing this has some strange side-effect
+            //  (probably because the pipe's in_active flag is true until a read 
+            //  is attempted)
+            rc = receive_and_process_zap_reply ();
+            if (rc == -1)
+                return -1;
+        } else if (!options.zap_enforce_domain) {
+            //  This supports the Stonehouse pattern (encryption without
+            //  authentication) in legacy mode (domain set but no handler).
+            state = sending_ready;
+        } else {
+            session->get_socket ()->event_handshake_failed_no_detail (
+              session->get_endpoint (), EFAULT);
             return -1;
+        }
+    } else {
+        //  This supports the Stonehouse pattern (encryption without authentication).
+        state = sending_ready;
     }
-    else
-        state = send_ready;
 
     return parse_metadata (initiate_plaintext + crypto_box_ZEROBYTES + 128,
                            clen - crypto_box_ZEROBYTES - 128);
@@ -513,33 +426,33 @@ int zmq::curve_server_t::process_initiate (msg_t *msg_)
 
 int zmq::curve_server_t::produce_ready (msg_t *msg_)
 {
+    const size_t metadata_length = basic_properties_len ();
     uint8_t ready_nonce [crypto_box_NONCEBYTES];
-    uint8_t ready_plaintext [crypto_box_ZEROBYTES + 256];
-    uint8_t ready_box [crypto_box_BOXZEROBYTES + 16 + 256];
+
+    uint8_t *ready_plaintext =
+      (uint8_t *) malloc (crypto_box_ZEROBYTES + metadata_length);
+    alloc_assert (ready_plaintext);
 
     //  Create Box [metadata](S'->C')
     memset (ready_plaintext, 0, crypto_box_ZEROBYTES);
     uint8_t *ptr = ready_plaintext + crypto_box_ZEROBYTES;
 
-    //  Add socket type property
-    const char *socket_type = socket_type_string (options.type);
-    ptr += add_property (ptr, "Socket-Type", socket_type, strlen (socket_type));
-
-    //  Add identity property
-    if (options.type == ZMQ_REQ
-    ||  options.type == ZMQ_DEALER
-    ||  options.type == ZMQ_ROUTER)
-        ptr += add_property (ptr, "Identity", options.identity, options.identity_size);
-
+    ptr += add_basic_properties (ptr, metadata_length);
     const size_t mlen = ptr - ready_plaintext;
 
     memcpy (ready_nonce, "CurveZMQREADY---", 16);
     put_uint64 (ready_nonce + 16, cn_nonce);
 
-    int rc = crypto_box_afternm (ready_box, ready_plaintext,
-                                 mlen, ready_nonce, cn_precom);
+    uint8_t *ready_box =
+      (uint8_t *) malloc (crypto_box_BOXZEROBYTES + 16 + metadata_length);
+    alloc_assert (ready_box);
+
+    int rc = crypto_box_afternm (ready_box, ready_plaintext, mlen, ready_nonce,
+                                 cn_precom);
     zmq_assert (rc == 0);
 
+    free (ready_plaintext);
+
     rc = msg_->init_size (14 + mlen - crypto_box_BOXZEROBYTES);
     errno_assert (rc == 0);
 
@@ -551,6 +464,7 @@ int zmq::curve_server_t::produce_ready (msg_t *msg_)
     //  Box [metadata](S'->C')
     memcpy (ready + 14, ready_box + crypto_box_BOXZEROBYTES,
             mlen - crypto_box_BOXZEROBYTES);
+    free (ready_box);
 
     cn_nonce++;
 
@@ -559,164 +473,20 @@ int zmq::curve_server_t::produce_ready (msg_t *msg_)
 
 int zmq::curve_server_t::produce_error (msg_t *msg_) const
 {
+    const size_t expected_status_code_length = 3;
     zmq_assert (status_code.length () == 3);
-    const int rc = msg_->init_size (6 + 1 + status_code.length ());
+    const int rc = msg_->init_size (6 + 1 + expected_status_code_length);
     zmq_assert (rc == 0);
     char *msg_data = static_cast <char *> (msg_->data ());
     memcpy (msg_data, "\5ERROR", 6);
-    msg_data [6] = sizeof status_code;
-    memcpy (msg_data + 7, status_code.c_str (), status_code.length ());
+    msg_data [6] = expected_status_code_length;
+    memcpy (msg_data + 7, status_code.c_str (), expected_status_code_length);
     return 0;
 }
 
 void zmq::curve_server_t::send_zap_request (const uint8_t *key)
 {
-    int rc;
-    msg_t msg;
-
-    //  Address delimiter frame
-    rc = msg.init ();
-    errno_assert (rc == 0);
-    msg.set_flags (msg_t::more);
-    rc = session->write_zap_msg (&msg);
-    errno_assert (rc == 0);
-
-    //  Version frame
-    rc = msg.init_size (3);
-    errno_assert (rc == 0);
-    memcpy (msg.data (), "1.0", 3);
-    msg.set_flags (msg_t::more);
-    rc = session->write_zap_msg (&msg);
-    errno_assert (rc == 0);
-
-    //  Request ID frame
-    rc = msg.init_size (1);
-    errno_assert (rc == 0);
-    memcpy (msg.data (), "1", 1);
-    msg.set_flags (msg_t::more);
-    rc = session->write_zap_msg (&msg);
-    errno_assert (rc == 0);
-
-    //  Domain frame
-    rc = msg.init_size (options.zap_domain.length ());
-    errno_assert (rc == 0);
-    memcpy (msg.data (), options.zap_domain.c_str (), options.zap_domain.length ());
-    msg.set_flags (msg_t::more);
-    rc = session->write_zap_msg (&msg);
-    errno_assert (rc == 0);
-
-    //  Address frame
-    rc = msg.init_size (peer_address.length ());
-    errno_assert (rc == 0);
-    memcpy (msg.data (), peer_address.c_str (), peer_address.length ());
-    msg.set_flags (msg_t::more);
-    rc = session->write_zap_msg (&msg);
-    errno_assert (rc == 0);
-
-    //  Identity frame
-    rc = msg.init_size (options.identity_size);
-    errno_assert (rc == 0);
-    memcpy (msg.data (), options.identity, options.identity_size);
-    msg.set_flags (msg_t::more);
-    rc = session->write_zap_msg (&msg);
-    errno_assert (rc == 0);
-
-    //  Mechanism frame
-    rc = msg.init_size (5);
-    errno_assert (rc == 0);
-    memcpy (msg.data (), "CURVE", 5);
-    msg.set_flags (msg_t::more);
-    rc = session->write_zap_msg (&msg);
-    errno_assert (rc == 0);
-
-    //  Credentials frame
-    rc = msg.init_size (crypto_box_PUBLICKEYBYTES);
-    errno_assert (rc == 0);
-    memcpy (msg.data (), key, crypto_box_PUBLICKEYBYTES);
-    rc = session->write_zap_msg (&msg);
-    errno_assert (rc == 0);
-}
-
-int zmq::curve_server_t::receive_and_process_zap_reply ()
-{
-    int rc = 0;
-    msg_t msg [7];  //  ZAP reply consists of 7 frames
-
-    //  Initialize all reply frames
-    for (int i = 0; i < 7; i++) {
-        rc = msg [i].init ();
-        errno_assert (rc == 0);
-    }
-
-    for (int i = 0; i < 7; i++) {
-        rc = session->read_zap_msg (&msg [i]);
-        if (rc == -1)
-            break;
-        if ((msg [i].flags () & msg_t::more) == (i < 6? 0: msg_t::more)) {
-            //  Temporary support for security debugging
-            puts ("CURVE I: ZAP handler sent incomplete reply message");
-            errno = EPROTO;
-            rc = -1;
-            break;
-        }
-    }
-
-    if (rc != 0)
-        goto error;
-
-    //  Address delimiter frame
-    if (msg [0].size () > 0) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: ZAP handler sent malformed reply message");
-        errno = EPROTO;
-        rc = -1;
-        goto error;
-    }
-
-    //  Version frame
-    if (msg [1].size () != 3 || memcmp (msg [1].data (), "1.0", 3)) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: ZAP handler sent bad version number");
-        errno = EPROTO;
-        rc = -1;
-        goto error;
-    }
-
-    //  Request id frame
-    if (msg [2].size () != 1 || memcmp (msg [2].data (), "1", 1)) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: ZAP handler sent bad request ID");
-        errno = EPROTO;
-        rc = -1;
-        goto error;
-    }
-
-    //  Status code frame
-    if (msg [3].size () != 3) {
-        //  Temporary support for security debugging
-        puts ("CURVE I: ZAP handler rejected client authentication");
-        errno = EACCES;
-        rc = -1;
-        goto error;
-    }
-
-    //  Save status code
-    status_code.assign (static_cast <char *> (msg [3].data ()), 3);
-
-    //  Save user id
-    set_user_id (msg [5].data (), msg [5].size ());
-
-    //  Process metadata frame
-    rc = parse_metadata (static_cast <const unsigned char*> (msg [6].data ()),
-                         msg [6].size (), true);
-
-error:
-    for (int i = 0; i < 7; i++) {
-        const int rc2 = msg [i].close ();
-        errno_assert (rc2 == 0);
-    }
-
-    return rc;
+    zap_client_t::send_zap_request ("CURVE", 5, key, crypto_box_PUBLICKEYBYTES);
 }
 
 #endif
